/* Binning with relative positions.
 * This version also keeps min/max scores per bin to speed up searching.
 *
 * As per binning2, but use arrays insetad of pointers for bin numbers
 * and range lists. This reduces memory layout and makes it amenable to
 * storage / retrieval from disk.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "array.h"
#include "binning.h"

/*
 * FIXME: this is shared between g_view, g_index and binning. It's a bit
 * back to front currently with store containing IO, but swapping this
 * around with io containing store will mean a common store struct that works
 * for all files.
 */
typedef struct bin_store {
    bin_index_t *(*get_bin)(struct bin_store *store, int bnum);
    void (*free_bin)(struct bin_store *store, int bnum);
    void *client_data;
} bin_store_t;

/* #define DISPLAY_MATCHES */
/* #define DEBUG */
#define MIN_BIN_SIZE 64

/*
 * Allocates and returns next range from a range array
 * Returns -1 for error.
 */
int next_range(Array ra) {
    if (NULL == ArrayRef(ra, ArrayMax(ra)))
	return -1;
    
    return ArrayMax(ra)-1;
}

/*
 * Simulate loading from disk.
 */
void load_range(bin_index_t *bin) {
    if (!bin->rng) {
	bin->rng = ArrayCreate(sizeof(range_t), 0);
    }
}

/*
 * Allocates and returns a bin number
 * Returns -1 for error
 */
int next_bin(Array ba) {
    if (NULL == ArrayRef(ba, ArrayMax(ba)))
	return -1;

    return ArrayMax(ba)-1;
}

/*
 * The recursive part of allocate_bin.
 * Returns the current bin allocated (ie ultimately the root bin_id
 */
int allocate_bins2(Array ba, int pos, int sz, int min_size) {
    int id = next_bin(ba);
    bin_index_t *bin = arrp(bin_index_t, ba, id);

    bin->pos = pos;
    bin->size = sz;
    bin->start_used = 0;
    bin->end_used = 0;
    bin->child[0] = 0;
    bin->child[1] = 0;
    bin->bin_id = id;
    /* bin->rng = NULL; */
    bin->rng = ArrayCreate(sizeof(range_t), 0);

    if (sz > min_size) {
	int lb = allocate_bins2(ba, 0, sz/2, min_size);
	int rb = allocate_bins2(ba, sz/2, sz - sz/2, min_size);
	
	bin = arrp(bin_index_t, ba, id); /* maybe diff to before */
	bin->child[0] = lb;
	bin->child[1] = rb;
    }

    return id;
}


/*
 * Recursively allocates the bin structure, halving bins until min_size
 * is reached.
 *
 * Returns an Array holding bin_index_t objects.
 */
Array allocate_bins(int pos, int sz, int min_size) {
    Array ba = ArrayCreate(sizeof(bin_index_t), 8);

    allocate_bins2(ba, pos, sz, min_size);
    return ba;
}

void dump_bins2(Array ba, int bin_num, int level, int offset) {
    int n;
    bin_index_t *bin = arrp(bin_index_t, ba, bin_num);

    offset += bin->pos;
    printf("%d: %d Bin, %06d-%06d (%06d-%06d used) (+%05d): l=%d r=%d:",
	   bin->bin_id, level, bin->pos, bin->pos + bin->size-1,
	   bin->start_used, bin->end_used, offset,
	   bin->child[0], bin->child[1]);
    load_range(bin);
    for (n = 0; n < ArrayMax(bin->rng); n++) {
	range_t *r = arrp(range_t, bin->rng, n);
	printf("\n\t%d-%d", r->start + offset, r->end + offset);
    }
    printf(" => %d objs\n", n);

    if (bin->child[0])
	dump_bins2(ba, bin->child[0], level+1, offset);
    if (bin->child[1])
	dump_bins2(ba, bin->child[1], level+1, offset);
}

void dump_bins(Array ba) {
    dump_bins2(ba, 0, 0, 0);
}

/*
 * Finds a bin that *entirely* contains the range start..end.
 * It also extends start_used and end_used optimisers.
 */
bin_index_t *find_bin(Array ba, int start, int end, int *offsetp) {
    bin_index_t *bin = arrp(bin_index_t, ba, 0); /* root node */
    int i, offset = bin->pos;

    /* Hacky and slow approach to start with */
    for (i = 0; i < 2 && bin->child[i];) {
	bin_index_t *ch = arrp(bin_index_t, ba, bin->child[i]);
	if (start >= ch->pos + offset &&
	    end <= ch->pos + offset + ch->size-1) {
	    bin = ch;
	    offset += bin->pos;
	    i = 0;
	} else {
	    i++;
	}
    }

    /* Adjust pos_used/size_used */
    start -= offset;
    end   -= offset;
    if (bin->start_used != bin->end_used) {
	if (bin->start_used > start)
	    bin->start_used = start;
	if (bin->end_used < end)
	    bin->end_used = end;
    } else {
	bin->start_used = start;
	bin->end_used = end;
    }

    /*
    printf("looking for %d..%d found bin %d..%d\n",
	   start, end, bin->pos, bin->pos+bin->size-1);
    */

    *offsetp = offset;
    
    return bin;
}

index_t *index_ranges(ranges_t *ranges, unsigned int glen) {
    index_t *idx = (index_t *)malloc(sizeof(*idx));
    int i;
    range_t *items = ranges ? ranges->ranges : NULL;
    int nitems = ranges ? ranges->nranges : 0;
    int off;
    Array ba;

    idx->rng = ranges;
    idx->glen = ++glen;

#if 0
    /* Round genome size to next power of 2 */
    glen--;
    glen |= glen>>1;
    glen |= glen>>2;
    glen |= glen>>4;
    glen |= glen>>8;
    glen |= glen>>16;
    glen++;
#endif

    /* Construct bin layout */
    fprintf(stderr, "glen = %d\n", glen);
    ba = allocate_bins(0, glen, MIN_BIN_SIZE);
    idx->data = ba;

    /* Index */
    for (i = 0; i < nitems; i++) {
	range_t *r;
	bin_index_t *bin = find_bin(ba, items[i].start, items[i].end, &off);
	int nr;

	items[i].start -= off;
	items[i].end -= off;

	if (!bin->rng)
	    bin->rng = ArrayCreate(sizeof(range_t), 0);

	nr = next_range(bin->rng);
	r = arrp(range_t, bin->rng, nr);
	*r = items[i]; /* struct copy */
    }

#ifdef DEBUG
    dump_bins(ba);
#endif

    return idx;
}

/* Add items to an existing index */
void index_range_add(index_t *idx, range_t *items, int nitems) {
    int i, off;
    Array ba = (Array)idx->data;

    for (i = 0; i < nitems; i++) {
	range_t *r;
	bin_index_t *bin;
	int nr;

	bin = find_bin(ba, items[i].start, items[i].end, &off);
	items[i].start -= off;
	items[i].end -= off;

	if (!bin->rng)
	    bin->rng = ArrayCreate(sizeof(range_t), 0);

	nr = next_range(bin->rng);
	r = arrp(range_t, bin->rng, nr);

	*r = items[i]; /* struct copy */
    }
}

static int nsearches = 0;
int count_items_in_range2(range_t *items, Array ba, int bin_num,
			  int start, int end, int offset) {
    int count = 0;
    bin_index_t *bin = arrp(bin_index_t, ba, bin_num);
    int i, n;
    range_t *l;

#ifdef DEBUG
    printf("%d looking for %d..%d found bin %d..%d(+%d)\n",
	   bin->bin_id, start, end, 
	   bin->pos, bin->pos+bin->size-1, offset);

    printf("Bin start/end is %d..%d (%d..%d with offset)\n",
	   bin->start_used, bin->end_used,
	   bin->start_used + offset, bin->end_used+ offset);
#endif
    /* FIXME: Separate to "items overlapping X-Y" and "items contained within
     * X-Y" functions. The former needs to check all buckets while the latter
     * can be optimised to only look in the buckets of the correct size.
     * Ie: an optimisation here is that we know an all objects in bin X with
     * children Y and Z will only contain items that happen to span both
     * Y and Z. Therefore when searching we do not need to check this
     * bin if we observe that the object *should* fit entirely within
     * a child bin.
     */

    /* Optimise - reject bin if query is outside of range of used objects */
    /* FIXME: Hack with goto to test theory ;-) */
    if (!(end < bin->start_used + offset || start > bin->end_used + offset)) {
	load_range(bin);
	for (n = 0; n < ArrayMax(bin->rng); n++) {
	    l = arrp(range_t, bin->rng, n);
	    nsearches++;
	    if (l->end + offset >= start
		&& l->start + offset <= end) {
#ifdef DISPLAY_MATCHES
		printf("Search %d..%d found %d..%d\n",
		       start, end, l->start + offset,
		       l->end + offset);
#endif
		count++;
	    } else {
#ifdef DISPLAY_MATCHES
		printf("Search %d..%d rejected %d..%d\n",
		       start, end, l->start + offset,
		       l->end + offset);
#endif
	    }
	}
    }

    /* Recurse down bins */
    for (i = 0; i < 2 && bin->child[i]; i++) {
	bin_index_t *ch = arrp(bin_index_t, ba, bin->child[i]);
#ifdef DEBUG
	printf("bin[%d]={%d+%d}+%d vs {%d,%d}\n",
	       i, ch->pos, ch->size,
	       offset, start, end);
#endif

	if (end   >= ch->pos + offset &&
	    start <= ch->pos + offset + ch->size-1) {
	    count += count_items_in_range2(NULL, ba, bin->child[i],
					   start, end, offset + ch->pos);
	} else if (ch->pos > end) {
	    break;
	}
    }

    return count;
}

int count_items_in_range(range_t *items, Array ba, int start, int end)
{
    return count_items_in_range2(items, ba, 0, start, end, 0);
}

int items_in_range2(bin_store_t *store, int bin_num, int start, int end,
		    int offset, range_t **results, int *alloc, int used) {
    int count = used;
    bin_index_t *bin = store->get_bin(store, bin_num);
    int i, n;
    range_t *l;

    if (!(end < bin->start_used + offset || start > bin->end_used + offset)) {
	load_range(bin);
	for (n = 0; n < ArrayMax(bin->rng); n++) {
	    l = arrp(range_t, bin->rng, n);
	    if (l->end + offset >= start
		&& l->start + offset <= end) {
		if (count >= *alloc) {
		    *alloc = *alloc ? *alloc * 2 : 16;
		    *results = (range_t *)realloc(*results,
						  *alloc * sizeof(range_t));
		}
		(*results)[count] = *l;
		(*results)[count].start += offset;
		(*results)[count].end   += offset;
		count++;
	    }
	}
    }

    /* Recurse down bins */
    for (i = 0; i < 2 && bin->child[i]; i++) {
	bin_index_t *ch = store->get_bin(store, bin->child[i]);
	if (end   >= ch->pos + offset &&
	    start <= ch->pos + offset + ch->size-1) {
	    count = items_in_range2(store, bin->child[i], start, end,
				    offset + ch->pos,
				    results, alloc, count);
	} else if (ch->pos > end) {
	    store->free_bin(store, bin->child[i]);
	    break;
	}
	store->free_bin(store, bin->child[i]);
    }

    store->free_bin(store, bin_num);
    return count;
}

range_t *items_in_range(bin_store_t *store, int start, int end, int *count)
{
    range_t *r = NULL;
    int alloc = 0;

    *count = items_in_range2(store, 0, start, end, 0, &r, &alloc, 0);
    return r;
}

static void (*indel_callback_func)(range_t *r, int pos, int dir) = NULL;

void indel_callback(void (*func)(range_t *r, int pos, int dir)) {
    indel_callback_func = func;
}

/*
 * Inserting to a bin requires the usual adjustment of items within that
 * bin (moving or extending) along with adjustments to the size or pos
 * fields of other bins.
 *
 * Specifically the insert will always take place in a leaf node. From here
 * increase the 'size' of the parent node and if we are the left child of
 * that parent increase the 'pos' of the right child. Then move up to that
 * parent node and repeat.
 *
 * This has the impact of correctly maintaining the relative 'offset' fields
 * computed elsewhere and so essentially moves all of the data right by 1
 * in the remainder of the tree.
 */
void insert_base2(Array ba, int bin_num, int pos, int offset) {
    range_t *l;
    int i, n;
    bin_index_t *bin = arrp(bin_index_t, ba, bin_num);

    /* Perform the insert into objects */
    if (pos <= bin->end_used + offset) {
	load_range(bin);
	for (n = 0; n < ArrayMax(bin->rng); n++) {
	    l = arrp(range_t, bin->rng, n);
	    if (l->end + offset >= pos &&
		l->start + offset <= pos) {
		if (indel_callback_func)
		    indel_callback_func(l, pos-offset, +1);
		/* Overlap => insert */
		l->end++;
	    } else if (l->start + offset > pos) {
		/* Move */
		l->start++;
		l->end++;
	    }
	}
    }

    /* Adjust the bin start/end/size positions too */
    bin->size++;
    if (pos < bin->start_used + offset)
	bin->start_used++;
    if (pos <= bin->end_used + offset)
	bin->end_used++;

    /* Recurse */
    if (bin->child[0] || bin->child[1]) {
	bin_index_t *ch;

	/* Find the correct child node */
       	for (i = 0; i < 2 && bin->child[i]; i++) {
	    ch = arrp(bin_index_t, ba, bin->child[i]);
	    if (pos >= ch->pos + offset &&
		pos <= ch->pos + offset + ch->size-1) {
		break;
	    }
	}

	/* Recurse */
	if (i < 2) {
	    insert_base2(ba, bin->child[i], pos, offset + ch->pos);
	}

	/* All children to the right of this one need pos updating too */
	if (i == 0) {
	    ch = arrp(bin_index_t, ba, bin->child[1]);
	    ch->pos++;
	}
    }
}

void insert_base(index_t *idx, int pos) {
#ifdef DEBUG
    printf("Insert to pos %d\n", pos);
#endif
    insert_base2((Array)idx->data, 0, pos, 0);
}

/*
 * Essentially this is the reverse of the insert_base2 function.
 * In order to keep the bin structure up to date we apply the same recursive
 * bucket 'move' procedure as used in insert_base2, but decrementing size
 * or position instead.
 */
void delete_base2(Array ba, int bin_num, int pos, int offset) {
    range_t *l;
    int i, n;
    bin_index_t *bin = arrp(bin_index_t, ba, bin_num);

    /* Perform the deletion from objects */
    if (pos <= bin->end_used + offset) {
	load_range(bin);
	for (n = 0; n < ArrayMax(bin->rng); n++) {
	    l = arrp(range_t, bin->rng, n);
	    if (l->end + offset >= pos &&
		l->start + offset <= pos) {
		if (indel_callback_func)
		    indel_callback_func(l, pos-offset, -1);
		/* Overlap => delete */
		l->end--;
	    } else if (l->start + offset > pos) {
		/* Move */
		l->start--;
		l->end--;
	    }
	}
    }

    /* Adjust the bin start/end/size positions too */
    bin->size--;
    if (pos < bin->start_used + offset)
	bin->start_used--;
    if (pos <= bin->end_used + offset)
	bin->end_used--;

    /* Recurse */
    if (bin->child[0] || bin->child[1]) {
	bin_index_t *ch;

	/* Find the correct child node */
       	for (i = 0; i < 2 && bin->child[i]; i++) {
	    ch = arrp(bin_index_t, ba, bin->child[i]);
	    if (pos >= ch->pos + offset &&
		pos <= ch->pos + offset + ch->size-1) {
		break;
	    }
	}

	/* Recurse */
	if (i < 2) {
	    delete_base2(ba, bin->child[i], pos, offset + ch->pos);
	}

	/* All children to the right of this one need pos updating too */
	if (i == 0) {
	    ch = arrp(bin_index_t, ba, bin->child[1]);
	    ch->pos--;
	}
    }
}

void delete_base(index_t *idx, int pos) {
#ifdef DEBUG
    printf("Delete from pos %d\n", pos);
#endif
    delete_base2((Array)idx->data, 0, pos, 0);
}


/*
 * Saves the bin layout and it's contents - cheap and cheerful, endian
 * specific, etc. It's just a hack to test the theory :-)
 * Format:
 * Bin details (pos, size, n-sub-bins (N), bin_id, item count (I)
 * I * item details
 * N * sub-bin details (ie recursively stored)
 *
 * It is expected that obj_save and the corresponding obj_load knows
 * how long the data will be either via implicit knowledge of by
 * writing/reading the object size.
 *
 * Returns 0 for success
 *        -1 for error
 */
int bin_save2(Array ba, int bnum, FILE *fp,
	      int (*obj_save)(FILE *fp, range_t *rng)) {
    int err = 0, i, nitems;
    bin_index_t *bin = arrp(bin_index_t, ba, bnum);

    /* The bin layout */
    fwrite(&bin->pos,        4, 1, fp);
    fwrite(&bin->size,       4, 1, fp);
    fwrite(&bin->start_used, 4, 1, fp);
    fwrite(&bin->end_used,   4, 1, fp);
    fwrite(&bin->bin_id,     4, 1, fp);
    fwrite(&bin->child[0],   4, 1, fp);
    fwrite(&bin->child[1],   4, 1, fp);

    /* Write items in this bin */
    nitems = bin->rng ? ArrayMax(bin->rng) : 0;
    fwrite(&nitems, 4, 1, fp);
    for (i = 0; i < nitems; i++) {
	err != obj_save(fp, arrp(range_t, bin->rng, i));
    }

    /* Now recurse */
    for (i = 0; i < 2; i++) {
	if (bin->child[i])
	    err |= bin_save2(ba, bin->child[i], fp, obj_save);
    }

    return err;
}

/*
 * Saves the bin layout and contents as binary data to 'fp'.
 * As this library doesn't understand the format of the data it does this
 * by calling the obj_save function ptr.
 *
 * Returns 0 for success
 *        -1 for error
 */
int bin_save(index_t *idx, FILE *fp,
	     int (*obj_save)(FILE *fp, range_t *rng)) {
    Array ba = (Array)idx->data;
    return bin_save2(ba, 0, fp, obj_save);
}


/*
 * The opposite of bin_save2().
 * This function loads the bin data file fp.
 *
 * Returns 0 for success
 *        -1 for error
 */
int bin_load2(Array ba, int bnum, FILE *fp,
	      int (*obj_load)(FILE *fp, range_t *rng)) {
    int err = 0, i, nitems;
    bin_index_t *bin;

    ArrayRef(ba, bnum);
    bin = arrp(bin_index_t, ba, bnum);

    /* The bin layout */
    fread(&bin->pos,        4, 1, fp);
    fread(&bin->size,       4, 1, fp);
    fread(&bin->start_used, 4, 1, fp);
    fread(&bin->end_used,   4, 1, fp);
    fread(&bin->bin_id,     4, 1, fp);
    fread(&bin->child[0],   4, 1, fp);
    fread(&bin->child[1],   4, 1, fp);

    /* Write items in this bin */
    fread(&nitems, 4, 1, fp);
    bin->rng = ArrayCreate(sizeof(range_t), nitems);
    for (i = 0; i < nitems; i++) {
	ArrayRef(bin->rng, i);
	err != obj_load(fp, arrp(range_t, bin->rng, i));
    }

    /* Now recurse */
    for (i = 0; i < 2; i++) {
	bin = arrp(bin_index_t, ba, bnum);
	if (bin->child[i]) {
	    err |= bin_load2(ba, bin->child[i], fp, obj_load);
	}
    }

    return err;
}


/*
 * The opposite of bin_save :-)
 *
 * Returns 0 for success
 *        -1 for error
 */
int bin_load(index_t *idx, FILE *fp,
	     int (*obj_load)(FILE *fp, range_t *rng)) {
    int err;
    Array ba = ArrayCreate(sizeof(bin_index_t), 0);
    bin_index_t *bin;

    err = bin_load2(ba, 0, fp, obj_load);

    idx->data = ba;
    bin = arrp(bin_index_t, ba, 0);
    idx->glen = bin->pos + bin->size;

    return err;
}
