#ifndef _HACHE_TABLE_H_
#define _HACHE_TABLE_H_

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

/* The data referenced by the hash table */
typedef union {
    uint64_t i;
    void *p;
} HashData;

/* A hash item with "next" pointer to use in a linked list */
typedef struct HashItemStruct {
    struct HashItemStruct *next;
    HashData data;        /* user defined data attached to this key */
    char    *key;         /* key we hashed on */
    int      key_len;     /* and its length */
    int      order;	  /* Index into HashTable->ordering[] */
    int	     ref_count;	  /* Starts as 1 */
} HashItem;

typedef struct HashOrderStruct {
    HashItem *hi;
    int next;
    int prev;
} HashOrder;

/* The main hash table structure itself */
typedef struct {
    int	      cache_size; /* Max size before purging cached entries */
    int       options;  /* HASH_FUNC & HASH_OPT macros */
    uint32_t  nbuckets; /* Number of hash buckets; power of 2 */
    uint32_t  mask;	/* bit-mask equiv of nbuckets */
    int       nused;    /* How many hash entries we're storing */
    HashItem **bucket;  /* The bucket "list heads" themselves */

    /* Cyclic cache array */
    HashOrder *ordering; 
    int head, tail, free;

    /* Cache I/O callback functions */
    void      *clientdata;
    HashData *(*load)(void *clientdata, char *key, int key_len);
    void      (*del)(void *clientdata, HashData data);

    int	      searches; /* number of total queries */
    int       hits;	/* number of cached queries */
} HashTable;

/* Functions to to use HashTable.options */
#define HASH_FUNC_HSIEH       0
#define HASH_FUNC_TCL         1
#define HASH_FUNC_JENKINS     2
#define HASH_FUNC_MASK        7

/* Other HashTable.options values */
#define HASH_NONVOLATILE_KEYS (1<<3)
#define HASH_ALLOW_DUP_KEYS   (1<<4)
#define HASH_DYNAMIC_SIZE     (1<<5)
#define HASH_OWN_KEYS	      (1<<6)

/* Hashing prototypes */
uint32_t hash(int func, uint8_t *key, int key_len);
uint32_t HashJenkins(uint8_t *k, int length);
uint32_t HashTcl(uint8_t *data, int len);
uint32_t HashHsieh(uint8_t *k, int length);

/* HashTable management prototypes */
HashTable *HashTableCreate(int size, int options);
void HashTableDestroy(HashTable *h, int deallocate_date);
int HashTableResize(HashTable *h, int newsize);
HashItem *HashTableAdd(HashTable *h, char *key, int key_len,
		       HashData data, int *new);
int HashTableDel(HashTable *h, HashItem *hi, int deallocate_data);
int HashTableRemove(HashTable *h, char *key, int key_len, int deallocate_data);
HashItem *HashTableSearch(HashTable *h, char *key, int key_len);
HashItem *HashTableNext(HashItem *hi, char *key, int key_len);
void HashTableDecRef(HashItem *hi);

void HashTablePurge(HashTable *h);
void HashTableStats(HashTable *h, FILE *fp);
void HashTableDump(HashTable *h, FILE *fp);

void HashOrderAccess(HashTable *h, HashItem *hi);
void HashOrderRemove(HashTable *h, HashItem *hi);
int HashOrderAdd(HashTable *h, HashItem *hi);

#endif /* _HACHE_TABLE_H_ */
