/*
 * Author:         James Bonfield, Feb 2007
 *                 Wellcome Trust Sanger Institute
 *
 * g_index: A program to index Tony Cox's .dat file format using a
 * recursive relative binning strategy. The data is stored in a g-library
 * format DB, for use with g_view for visualisation.
 *
 * The .dat alignment format consists of one line per sequence in order:
 *
 * seq_name ref_name start_pos end_pos direction(+1 or -1) clip_left \
 *    clip_right 0(?) 0(?) ? ? sequence confidence_values
 * 
 * Coordinates all start with 1 being the first base.
 */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include <g-struct.h>
#include <g-db.h>
#include <g-os.h>
#include <g-filedefs.h>
#include <g-error.h>
#include <g-defs.h>
#include <g-misc.h>
#include <g-io.h>
#include <g-request.h>
#include <g-connect.h>

#include "binning.h"
#include "array.h"
#include "tg_io.h"

/* Uncomment to read the old .dat format */
/* #define OLD_DAT */

/* ------------------------------------------------------------------------ */
/* Sequence manipulation - ripped out of Staden Package's seq_utils.c */

/*
 * Reverses and complements a piece of DNA
 */
void complement_seq(char *seq, char *conf, int seq_len) {
    int i, middle, j;
    char temp;
    static int complementary_base[256];
    static int init = 0;

    if (!init) {
	for (i = 0; i < 256; i++)
	    complementary_base[i] = i;

	complementary_base['a'] = 't';
	complementary_base['c'] = 'g';
	complementary_base['g'] = 'c';
	complementary_base['t'] = 'a';
	complementary_base['u'] = 'a';
	complementary_base['A'] = 'T';
	complementary_base['C'] = 'G';
	complementary_base['G'] = 'C';
	complementary_base['T'] = 'A';
	complementary_base['U'] = 'A';

	complementary_base['n'] = 'n';
	complementary_base['-'] = '-';
	complementary_base['b'] = 'v';
	complementary_base['d'] = 'h';
	complementary_base['h'] = 'd';
	complementary_base['k'] = 'm';
	complementary_base['m'] = 'k';
	complementary_base['r'] = 'y';
	complementary_base['s'] = 's';
	complementary_base['v'] = 'b';
	complementary_base['w'] = 'w';
	complementary_base['y'] = 'r';

	complementary_base['B'] = 'V';
	complementary_base['D'] = 'H';
	complementary_base['H'] = 'D';
	complementary_base['K'] = 'M';
	complementary_base['M'] = 'K';
	complementary_base['R'] = 'Y';
	complementary_base['S'] = 'S';
	complementary_base['V'] = 'B';
	complementary_base['W'] = 'W';
	complementary_base['Y'] = 'R';
	init = 1;
    }

    middle = seq_len/2;
    for ( i = 0, j = seq_len-1; i < middle; i++, j--) {
	temp = complementary_base [ (unsigned char) seq[i] ];
	seq[i] = complementary_base [ (unsigned char) seq[j] ];
	seq[j] = temp;
	temp = conf[i];
	conf[i] = conf[j];
	conf[j] = temp;
    }

    if ( seq_len % 2 )
      seq[middle] = complementary_base [ (unsigned char) seq[middle] ];
}

/* ------------------------------------------------------------------------ */
/*
 * File format reading code. We should add in here solexa eland support and
 * similar.
 */

/*
 * Parses a line of text and stores the data in seq.
 *
 * Returns 0 for success
 *        -1 for failure
 */
int parse_line(seq_t *s, char *line) {
    char name[1024], seq[1024], conf[1024];
    int start, end, dir, cleft, cright, ind, i;
    size_t len;

#ifdef OLD_DAT
#    define TEMPLATE "%s %*s %d %d %d %d %d %*d %*d %s %n"
#else
#    define TEMPLATE "%s %*s %d %d %d %d %d %*d %*d %*d %*d %s %n"
#endif

    /* Prevent valgrind from complaining about portions of seq being unset */
    memset(s, 0, sizeof(*s));

    /* Decode the line */
    if (sscanf(line, TEMPLATE,
               name, &start, &end, &dir, &cleft, &cright, seq, &ind) == 7) {
        len = strlen(seq);
        line = &line[ind];
        for (i = 0; i < len; i++) {
            conf[i] = strtol(line, &line, 10);
        }

        s->pos = start;
        s->len = dir * len;
        s->left = cleft;
        s->right = cright;

	/*
	 * FIXME: mark as complemented without actually complementing, we'll
	 * do this later on-the-fly in the viewer
	 */
        if (-1 == dir) {
            int tmp;
            complement_seq(seq, conf, len);
            tmp = s->left;
            s->left = len - (s->right-1);
            s->right = len - (tmp-1);
        }
	s->pos -= s->left-1;

	s->name_len = strlen(name);
	s->name = s->data = (char *)malloc(s->name_len+2*len);
	s->seq = s->data + s->name_len;
	s->conf = s->seq + len;
        strcpy(s->name, name);
        memcpy(s->seq, seq, len);
        memcpy(s->conf, conf, len);
        return 0;
    }

    return -1;
}

/*
 * Parses the .dat file passed in, writing sequences to 'io' and adding to
 * bins/ranges pointed to by 'index'.
 *
 * Returns 0 on success
 *	  -1 on error
 */
int parse_file(IO *io, index_t *index, int max_size, char *dat_fn) {
    int nseqs = 0;
    struct stat sb;
    FILE *dat_fp;
    unsigned char line[1024];
    off_t pos;

    printf("Loading %s...\n", dat_fn);
    if (-1 == stat(dat_fn, &sb) ||
	NULL == (dat_fp = fopen(dat_fn, "r"))) {
	perror(dat_fn);
	return -1;
    }

    /* Loop:
     * Read 1 sequence
     * Parse it & create a range
     * Save sequence
     * Insert to index
     */
    pos = 0;
    while (fgets(line, 1024, dat_fp)) {
	seq_t seq;
	range_t r;
	int recno;

	pos += strlen(line);

	if (-1 == parse_line(&seq, line)) {
	    fprintf(stderr, "Parser error on line: %s", line);
	    seq.len = 0;
	}

	/* Save sequence */
	recno = save_seq(io, &seq);
	free(seq.data);

	/* Create range */
	r.start = seq.pos;
	r.end   = seq.pos + (seq.len > 0 ? seq.len : -seq.len) - 1;
	r.rec   = recno;

	if (r.end > max_size) {
	    fprintf(stderr, "Position %d is greater than max. index size\n",
		    r.end);
	    nseqs++;
	    continue;
	}

	/*
	 * This is the real memory hog here. We're storing all bin and range
	 * data in memory and only writing at the end.
	 */
	index_range_add(index, &r, 1);

	nseqs++;
	if ((nseqs % 10000) == 0) {
	    static int perc = 0;
	    if (perc < 100.0 * pos / sb.st_size) {
		perc = 100.0 * pos / sb.st_size;
		printf("\r%d%%", perc);
		fflush(stdout);
	    }
	}
    }

    fclose(dat_fp);

    return 0;
}

/* ------------------------------------------------------------------------ */
/* 'Binning' callback or utilisation functions */

int write_bins2(IO *io, Array ba, int bnum, int rec, int *root) {
    GBin g;
    bin_index_t *bin = arrp(bin_index_t, ba, bnum);
    GView v;
    int err = 0;

    g.pos = bin->pos;
    g.size = bin->size;
    g.start = bin->start_used;
    g.end = bin->end_used;
    g.child[0] = bin->child[0] ? allocate(io) : 0;
    g.child[1] = bin->child[1] ? allocate(io) : 0;
    g.range = bin->rng ? allocate(io) : 0;
    g.id = bin->bin_id;

    /* Write GBin */
    v = lock(io, rec);
    err |= g_write_le4(io, v, &g, sizeof(g));
    err |= g_flush(io, v);
    err |= unlock(io, v);

    if (root)
	*root = rec;

    /* Write GRange */
    if (bin->rng) {
	v = lock(io, g.range);
	err |= g_write_le4(io, v, ArrayBase(range_t, bin->rng),
			   sizeof(range_t) * ArrayMax(bin->rng));
	err |= g_flush(io, v);
	err |= unlock(io,v);

	if (0) {
	    int i;
	    GViewInfo vi;

	    g_view_info_(io->db, io->client, v, &vi);
	    printf("\trange rec %d, ranges %d, size %d, err %d\n",
		   g.range, ArrayMax(bin->rng),
		   vi.used, err);

	    for (i = 0; i < ArrayMax(bin->rng); i++) {
		range_t *r = arrp(range_t, bin->rng, i);
		printf("\t%d\n", r->rec);
	    }
	}
    }

    if (bin->child[0])
	write_bins2(io, ba, bin->child[0], g.child[0], NULL);
    if (bin->child[1])
	write_bins2(io, ba, bin->child[1], g.child[1], NULL);

    return err;
}

int write_bins(IO *io, index_t *idx, int *root) {
    Array ba = (Array)idx->data;
    return write_bins2(io, ba, 0, allocate(io), root);
}

int save_seq(IO *io, seq_t *seq) {
    GView v;
    int rec;
    int err = 0;
    unsigned char block[1024], *copy = block;
    int sz = seq->name_len + 2*ABS(seq->len); /* name,seq,conf */
    int sz2 = sizeof(GSeq_header) + 1 + sz;   /* Header, name_len, sz.data */

    if (sz2 > 1024) {
	if (NULL == (copy = (char *)malloc(sz2)))
	    return 1;
    }

    rec = allocate(io);
    v = lock(io, rec);

    ((GSeq_header *)copy)->pos      = le_int4(seq->pos);
    ((GSeq_header *)copy)->len      = le_int4(seq->len);
    ((GSeq_header *)copy)->left     = le_int4(seq->left);
    ((GSeq_header *)copy)->right    = le_int4(seq->right);
    copy[sizeof(GSeq_header)]       =         seq->name_len;
    memcpy(copy + sizeof(GSeq_header) + 1, seq->data, sz);

    err |= g_write(io, v, (void *)copy, sz2);
    err |= g_flush(io, v);
    err |= unlock(io,v);

    if (copy != block)
	free(copy);

    return rec;
}

/* ------------------------------------------------------------------------ */
void usage() {
    fprintf(stderr, "Usage: g_index [-l length] [-o output] dat_file ...\n");
}

int main(int argc, char **argv) {
    index_t *index;
    unsigned int max_size = 1000000;
    IO *io;
    int root_bin;
    GContig c;
    int contig_rec;
    GView contig_view;
    int opt;
    char *out_fn = "g_db";

    printf("\n\tg_index:\tShort Read Alignment Indexer, version 0.3\n");
    printf("\n\tAuthor: \tJames Bonfield (jkb@sanger.ac.uk)\n");
    printf("\t        \t2007, Wellcome Trust Sanger Institute\n\n");

    /* Arg parsing */
    while ((opt = getopt(argc, argv, "hl:o:")) != -1) {
	switch(opt) {
	case 'l':
	    max_size = atoi(optarg);
	    break;

	case 'o':
	    out_fn = optarg;
	    break;

	case 'h':
	    usage();
	    return 0;

	default:
	    if (opt == ':')
		fprintf(stderr, "Missing parameter\n");
	    else
		fprintf(stderr, "Unknown option '%c'\n", opt);
	    usage();
	    return 1;
	}
    }

    if (optind == argc) {
	usage();
	return 1;
    }

    /* Open the DB */
    create_new_files(out_fn, 8, 100);
    io = g_open(out_fn, 0);

    /* Create the main GContig. record - hard coded to be record 0 */
    contig_rec = allocate(io);
    contig_view = lock(io, contig_rec);

    /* Create a dummy index, but we need to know the length upfront... */
    index = (index_t *)index_ranges(NULL, max_size);

    /* File processing loop */
    while (optind < argc) {
	parse_file(io, index, max_size, argv[optind++]);
    }

    /* Write bins & ranges */
    fprintf(stderr, "\nWriting bins / ranges\n");
    if (0 != write_bins(io, index, &root_bin)) {
	fprintf(stderr, "Error writing bins\n");
	return 1;
    }

    /* Write the master contig record */
    c.length = max_size;
    c.bin = root_bin;
    printf("rec %d, size %d\n", contig_rec, sizeof(c));
    g_write_le4(io, contig_view, &c, sizeof(c));
    g_flush(io, contig_view);
    unlock(io, contig_view);

    /* tree_print(io->db->gfile->freetree); */

    g_free(io);

    return 0;
}
