/*
 * Author:         James Bonfield, Feb 2007
 *                 Wellcome Trust Sanger Institute
 *
 * g_view:
 * Loads and displays the contents of a g-library format database as created
 * by the g_index application.
 */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <curses.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>

#include <g-struct.h>
#include <g-db.h>
#include <g-os.h>
#include <g-io.h>
#include <g-filedefs.h>
#include <g-error.h>
#include <g-defs.h>
#include <g-misc.h>
#include <g-request.h>
#include <g-connect.h>

#include "binning.h"
#include "array.h"
#include "tg_struct.h"
#include "tg_io.h"

/* ------------------------------------------------------------------------ */
/* Globals - yeah yeah, it's just a test hack! */

typedef struct bin_store {
    bin_index_t *(*get_bin)(struct bin_store *store, int bnum);
    void (*free_bin)(struct bin_store *store, int bnum);
    int root_bin;
    IO *io;
    HashTable *bin_hash;
    HashTable *seq_hash;
} bin_store_t;

/* ------------------------------------------------------------------------ */
/* Binning callback functions. These perform the nuts and bolts of I/O */

static bin_index_t *io_bin_load(IO *io, int bnum) {
    GBin b;
    GView v;
    bin_index_t *bin = (bin_index_t *)calloc(1, sizeof(*bin));

    /* Load GBin */
    v = lock(io, bnum);
    g_read_le4(io, v, &b, sizeof(b));
    unlock(io, v);

    /* Construct bin */
    bin->pos = b.pos;
    bin->size = b.size;
    bin->start_used = b.start;
    bin->end_used = b.end;
    bin->child[0] = b.child[0];
    bin->child[1] = b.child[1];
    bin->bin_id = b.id;
    bin->rng = NULL;

    /* Load ranges */
    if (b.range) {
	GViewInfo vi;
	int nranges;

	v = lock(io, b.range);
	g_view_info_(io->db, io->client, v, &vi);
	nranges = vi.used / sizeof(range_t);

	bin->rng = ArrayCreate(sizeof(range_t), nranges);
	ArrayRef(bin->rng, nranges-1);
	g_read_le4(io, v, ArrayBase(range_t, bin->rng),
		   nranges * sizeof(range_t));
	unlock(io, v);
    }

    return bin;
}


/* Callback from Hache */
static HashData *io_bin_load2(void *clientdata, char *key, int key_len) {
    bin_store_t *store = (bin_store_t *)clientdata;
    static HashData hd;
    int bnum = *(int *)key;

    hd.p = io_bin_load(store->io, bnum);
    return &hd;
}

bin_index_t *get_bin(bin_store_t *store, int bnum) {
    HashItem *hi;
    if (!bnum)
	bnum = store->root_bin;

    hi = HashTableSearch(store->bin_hash, (char *)&bnum, sizeof(bnum));
    return hi ? (bin_index_t *)hi->data.p : NULL;
}


/* Callback from Hache */
static void io_bin_free(void *clientdata, HashData hd) {
    bin_index_t *bin = hd.p;

    if (!bin)
	return;

    if (bin->rng)
	ArrayDestroy(bin->rng);

    free(bin);
}

void free_bin(bin_store_t *store, int bnum) {
    HashItem *hi;
    hi = HashTableSearch(store->bin_hash, (char *)&bnum, sizeof(bnum));
    if (hi) HashTableDecRef(hi);
}


/* Callback from Hache */
static HashData *io_seq_load2(void *clientdata, char *key, int key_len) {
    bin_store_t *store = (bin_store_t *)clientdata;
    static HashData hd;
    int rec = *(int *)key;
    GSeq_header *buf;
    seq_t *seq;
    size_t len;
    GView v;

    v = lock(store->io, rec);
    buf = g_read_alloc(store->io, v, &len);

    /* Layout = seq struct + data in one malloc segment */
    seq = (seq_t *)malloc(sizeof(seq_t) + len-sizeof(GSeq_header));
    seq->pos   = le_int4(buf->pos);
    seq->len   = le_int4(buf->len);
    seq->left  = le_int4(buf->left);
    seq->right = le_int4(buf->right);
    seq->name_len = ((unsigned char *)buf)[sizeof(GSeq_header)];
    memcpy(&seq->data, ((unsigned char *)buf) + sizeof(GSeq_header) + 1,
	   seq->name_len + 2*ABS(seq->len));
    seq->name = &seq->data;
    seq->seq = seq->name + seq->name_len;
    seq->conf = seq->seq + ABS(seq->len);
    unlock(store->io, v);

    free(buf);

    hd.p = seq;
    return &hd;
}

void free_seq(bin_store_t *store, int rec) {
    HashItem *hi = HashTableSearch(store->seq_hash, (char *)&rec, sizeof(rec));
    if (hi) HashTableDecRef(hi);
}

seq_t *get_seq(bin_store_t *store, int rec) {
    HashItem *hi = HashTableSearch(store->seq_hash, (char *)&rec, sizeof(rec));
    return hi ? (seq_t *)hi->data.p : NULL;
}

void store_destroy(bin_store_t *store) {
    if (store->io) {
	g_free(store->io);
    }

    if (store->bin_hash) {
	puts("\n=== bin_hash ===");
	HashTableStats(store->bin_hash, stdout);
	HashTableDestroy(store->bin_hash, 1);
    }

    if (store->seq_hash) {
	puts("\n=== seq_hash ===");
	HashTableStats(store->seq_hash, stdout);
	HashTableDestroy(store->seq_hash, 1);
    }

    free(store);
}

/* ------------------------------------------------------------------------ */
/* Consensus generation functions */

/* ACGT to 0123 conversion */
static unsigned char lookup[256], lookup_done = 0;

/*
 * Compute a basic non-weighted consensus. We simply pick the basecall
 * most frequently used.
 *
 * FIXME: use a weighted sum based on confidence values instead?
 */
int calc_cons(bin_store_t *store, range_t *r, int nr, int xpos, int wid,
	      char *cons) {
    int i, j;
    int (*cvec)[6] = (int (*)[6])calloc(wid, 6 * sizeof(int));

    if (!lookup_done) {
	memset(lookup, 5, 256);
	lookup_done = 1;
	lookup['A'] = lookup['a'] = 0;
	lookup['C'] = lookup['c'] = 1;
	lookup['G'] = lookup['g'] = 2;
	lookup['T'] = lookup['t'] = 3;
	lookup['*'] = lookup[','] = 4;
    }

    /* Accumulate */
    for (i = 0; i < nr; i++) {
	int sp = r[i].start;
	seq_t *s = get_seq(store, r[i].rec);
	int l = s->len > 0 ? s->len : -s->len;
	unsigned char *seq = (unsigned char *)s->seq;
	int left = s->left, right = s->right;

	if (sp < xpos) {
	    seq   += xpos - sp;
	    l     -= xpos - sp;
	    left  -= xpos - sp;
	    right -= xpos - sp;
	    sp = xpos;
	}
	if (l > wid - (sp-xpos))
	    l = wid - (sp-xpos);
	if (left < 1)
	    left = 1;

	for (j = left-1; j < right; j++) {
	    if (sp-xpos+j < wid)
		cvec[sp-xpos+j][lookup[seq[j]]]++;
	}
	free_seq(store, r[i].rec);
    }

    memset(cons, ' ', wid);

    /* and speculate :-) */
    for (i = 0; i < wid; i++) {
	int max, max_base = 5;
	for (max = j = 0; j < 6; j++) {
	    if (max < cvec[i][j]) {
		max = cvec[i][j];
		max_base = j;
	    }
	}
	cons[i] = "ACGT*N"[max_base];
    }

    free(cvec);

    return 0;

}

/* ------------------------------------------------------------------------ */
/* Curses display and keyboard handling routines */
static WINDOW *gotowin = NULL;
static WINDOW *helpwin = NULL;

#define DISPLAY_COLOURS 1
#define DISPLAY_CUTOFFS 2
#define DISPLAY_QUAL    4
#define DISPLAY_DIFFS   8

void exit_curses(int sig) {
    endwin();
    exit(0);
}

void init_curses() {
    signal(SIGINT, exit_curses);

    initscr();
    keypad(stdscr, TRUE);
    start_color();
    init_pair(0, COLOR_RED,    COLOR_BLACK);
    init_pair(1, COLOR_GREEN,  COLOR_BLACK);
    init_pair(2, COLOR_BLUE,   COLOR_BLACK);
    init_pair(3, COLOR_YELLOW, COLOR_BLACK);
    init_pair(4, COLOR_WHITE,  COLOR_BLACK);
    init_pair(5, COLOR_WHITE,  COLOR_BLACK);
    clear();
    noecho();
    cbreak();

    gotowin = newwin(3, 26, 10, 5);
    helpwin = newwin(18,40, 10, 5);
}

int wgoto(int xpos, WINDOW *win) {
    int orig = xpos;
    
    /* wborder(win, 0, 0, 0, 0, 0, 0, 0, 0); */
    wborder(win, '|', '|', '-', '-', '+', '+', '+', '+');

    mvwprintw(win, 1, 2, "Goto base:             ", xpos);
    mvwprintw(win, 1, 13, "%-d", xpos);

    for (;;) {
	int c;
	wrefresh(win);

	switch (c = wgetch(win)) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    xpos = xpos * 10 + c - '0';
	    break;

	case KEY_BACKSPACE:
	case '\010':
	case '\177':
	    xpos /= 10;
	    break;

	case '\027': /* control w */
	    xpos = 0;
	    break;

	case '\033': /* escape */
	    return orig;

	case KEY_ENTER:
	case '\012':
	case '\015':
	    return xpos;
	}

	mvwprintw(win, 1, 13, "            ");
	mvwprintw(win, 1, 13, "%-d", xpos);
    }

    return 0;
}

void whelp(WINDOW *win) {
    int r = 1;

    /* wborder(win, 0, 0, 0, 0, 0, 0, 0, 0); */
    wborder(win, '|', '|', '-', '-', '+', '+', '+', '+');

    mvwprintw(win, r++, 2, "        -=-    Help    -=- ");
    r++;
    mvwprintw(win, r++, 2, "?         This window");
    mvwprintw(win, r++, 2, "Arrows    Small scroll movement");
    mvwprintw(win, r++, 2, "h,j,k,l   Small scroll movement");
    mvwprintw(win, r++, 2, "H,J,K,L   Large scroll movement");
    mvwprintw(win, r++, 2, "ctrl-H    Scroll 1k left");
    mvwprintw(win, r++, 2, "ctrl-L    Scroll 1k right");
    mvwprintw(win, r++, 2, "c         Toggle cutoffs");
    mvwprintw(win, r++, 2, "C         Toggle colours");
    mvwprintw(win, r++, 2, "q         Toggle quality");
    mvwprintw(win, r++, 2, "d         Toggle differences");
    mvwprintw(win, r++, 2, "-         Decrement qual. threshold");
    mvwprintw(win, r++, 2, "+         Inccrement qual. threshold");
    mvwprintw(win, r++, 2, "g         Go to specific location");
    mvwprintw(win, r++, 2, "x         Exit");
    wrefresh(win);
    wgetch(win);
}

/* Sort comparison function for range_t; sort by ascending position */
static int sort_range(const void *v1, const void *v2) {
    const range_t *r1 = (const range_t *)v1;
    const range_t *r2 = (const range_t *)v2;
    return r1->start - r2->start;
}

static void display_gap(bin_store_t *store, int xpos, int ypos, int nlines,
			int wid, int mode, int qual_cutoff,
			int in_curses) {
    range_t *r;
    int i, nr, lno, y;
    char line[1024], *lp;
    char cons[1024];
    int attr;

    wid -= MAX_NAME_LEN+2;

    if (xpos < wid/2+2)
	xpos = wid/2+2;
    
    xpos -= wid/2;

    /* Query visible objects */
    r = (range_t *)items_in_range(store, xpos, xpos+wid-1, &nr);
    qsort(r, nr, sizeof(*r), sort_range);

    /* Consensus */
    calc_cons(store, r, nr, xpos, wid, cons);
    if (in_curses) {
	clear();
	mvaddnstr(0, MAX_NAME_LEN+2, cons, wid);
    } else {
	printf("%*s%.*s\n", MAX_NAME_LEN+2, "", wid, cons);
    }

    /* Position */
    for (lp = line, i = xpos; i < xpos+wid+19; i++) {
	if (i % 10 == 0) {
	    sprintf(lp, "%10d", i-10);
	    lp += 10;
	}
    }
    if (in_curses) {
	mvaddnstr(1, MAX_NAME_LEN+2, line+9+((xpos-1)%10), wid);
    } else {
	printf("%*s%.*s\n", MAX_NAME_LEN+2, "", wid,
	       line+9+((xpos-1)%10), wid);
    }


    /* Sequences */
    for (i = y = 0; i < nr && y < ypos; i++, y++);
    for (lno = 2; i < nr && lno < nlines; i++, lno++) {
	seq_t *s = get_seq(store, r[i].rec);
	int sp = r[i].start;
	int l = s->len > 0 ? s->len : -s->len;
	unsigned char seq_a[MAX_SEQ_LEN], *seq = seq_a;
	int j;
	int left = s->left, right = s->right;
	char *conf = s->conf;

	memcpy(seq, s->seq, l);

	if (sp < xpos) {
	    seq  += xpos - sp;
	    conf += xpos - sp;
	    l     -= xpos - sp;
	    left  -= xpos - sp;
	    right -= xpos - sp;
	    sp = xpos;
	}
	if (l > wid - (sp-xpos))
	    l = wid - (sp-xpos);

	if (in_curses) {
	    mvaddch(lno, 0, "-+"[s->len>=0]);
	    addnstr(s->name, s->name_len);
	    move(lno, MAX_NAME_LEN+2+sp-xpos);
	} else {
	    printf("%c%.*s%*s",
		   "-+"[s->len>=0],
		   MIN(MAX_NAME_LEN, s->name_len), s->name,
		   MAX_NAME_LEN+1-MIN(MAX_NAME_LEN, s->name_len) +sp-xpos, "");
	}

	for (j = 0; j < l; j++) {
	    attr = (mode & DISPLAY_COLOURS) ? COLOR_PAIR(lookup[seq[j]]) : 0;

	    if (mode & DISPLAY_DIFFS
		&& sp-xpos+j < wid && seq[j] == cons[sp-xpos+j])
		seq[j] = '.';
	    if (j < left-1 || j > right-1)
		seq[j] = (mode & DISPLAY_CUTOFFS) ? tolower(seq[j]) : ' ';

	    if (conf[j] >= qual_cutoff && mode & DISPLAY_QUAL) {
		attr |= A_BOLD;
	    }

	    if (in_curses) {
		addch(seq[j] | attr);
	    } else {
		putchar(seq[j]);
	    }
	}

	if (!in_curses)
	    putchar('\n');

	free_seq(store, r[i].rec);
    }

    if (in_curses)
	refresh();

    free(r);
}

void curses_loop(bin_store_t *store, int xpos, int mode) {
    int ypos = 0;
    int maxy, maxx;
    int qual_cutoff = 29;

#ifdef NCURSES_VERSION
    getmaxyx(stdscr, maxy, maxx);
#else
    maxx = 80;
    maxy = 40;
#endif

    if (xpos < (maxx-MAX_NAME_LEN+2)/2)
	xpos = (maxx-MAX_NAME_LEN+2)/2;

    display_gap(store, xpos, ypos, maxy, maxx, mode, qual_cutoff, 1);

    for(;;) {
	int c;

	switch(c = getch()) {
	case 'd':
	    mode ^= DISPLAY_DIFFS;
	    break;

	case 'c':
	    mode ^= DISPLAY_CUTOFFS;
	    break;

	case 'C':
	    mode ^= DISPLAY_COLOURS;
	    break;

	case 'q':
	    mode ^= DISPLAY_QUAL;
	    break;

	case 'g':
	    xpos = wgoto(xpos, gotowin);
	    break;

	case '?':
	    whelp(helpwin);
	    break;

	case '-':
	    qual_cutoff--;
	    break;

	case '+':
	    qual_cutoff++;
	    break;

	/* Orthoganol movements */
	case KEY_RIGHT:
	case 'l':
	    xpos++;
	    break;

	case KEY_SRIGHT:
	case 'L':
	    xpos+=20;
	    break;

	case '\014':
	    xpos+=1000;
	    break;

	case KEY_LEFT:
	case 'h':
	    xpos--;
	    break;

	case KEY_SLEFT:
	case 'H':
	    xpos-=20;
	    break;

	case '\010':
	    xpos-=1000;
	    break;

	case KEY_DOWN:
	case 'j':
	    ypos++;
	    break;

	case 'J':
	    ypos+=20;
	    break;

	case KEY_UP:
	case 'k':
	    ypos--;
	    break;

	case 'K':
	    ypos-=20;
	    break;

	case 'x':
	    return;

#ifdef KEY_RESIZE
	case KEY_RESIZE:
	    getmaxyx(stdscr, maxy, maxx);
	    break;
#endif

	default:
	    continue;
	}

	if (xpos < 0) xpos = 0;
	if (ypos < 0) ypos = 0;

	display_gap(store, xpos, ypos, maxy, maxx, mode, qual_cutoff, 1);
    }
}

/*
 * Prints output to stdout without any curses control. This is a distilled
 * down display_gap() function.
 */
print_output(bin_store_t *store, int xpos, int width, int mode) {
    display_gap(store, xpos, 0, 100000, width, mode, 0, 0);
}

/* ------------------------------------------------------------------------ */
/* Debug functions that don't use curses - handy for valgrind testing */
static void test_mode(bin_store_t *store, int xpos) {
    range_t *r;
    int nr, i;

    r = (range_t *)items_in_range(store, xpos, xpos+79, &nr);
    qsort(r, nr, sizeof(*r), sort_range);
    for (i = 0; i < nr; i++) {
	seq_t *s = get_seq(store, r[i].rec);
	printf("%d: %s\n", i, s->name);
	free_seq(store, r[i].rec);
    }

    store_destroy(store);
    system("ps lx | grep g_iotest | grep -v grep");
    exit(0);
}

static void test_mode2(bin_store_t *store, int xpos) {
    range_t *r;
    int nr;
    char cons[20];

    r = (range_t *)items_in_range(store, xpos-50, xpos+100, &nr);
    qsort(r, nr, sizeof(*r), sort_range);
    calc_cons(store, r, nr, xpos, 50, cons);
    printf("Cons=%.*s\n", 50, cons);
    store_destroy(store);
    exit(0);
}

/* ------------------------------------------------------------------------ */
void usage() {
    fprintf(stderr, "Usage: tg_view [options] dbname [position]\n");
    fprintf(stderr, "\t-h        This help\n");
    fprintf(stderr, "\t-d        Start in highlight-disagreements mode\n");
    fprintf(stderr, "\t-C        Start with cutoffs hidden\n");
    fprintf(stderr, "\t-c        Start with cutoffs shown (default)\n");
    fprintf(stderr, "\t-l width  Output non-interactively to stdout\n");
}


int main(int argc, char **argv) {
    IO *io;
    GView v;
    int xpos = 0;
    bin_store_t *store = (bin_store_t *)calloc(1, sizeof(*store));
    int opt;
    int lp_mode = 0;
    int mode = DISPLAY_QUAL | DISPLAY_CUTOFFS;
    extern char *optarg;

    while ((opt = getopt(argc, argv, "hl:dcC")) != -1) {
	switch (opt) {
	case 'h':
	    usage();
	    return 0;

	case 'd':
	    mode |= DISPLAY_DIFFS;
	    break;

	case 'c':
	    mode |= DISPLAY_CUTOFFS;
	    break;

	case 'C':
	    mode &= ~DISPLAY_CUTOFFS;
	    break;

	case 'l':
	    lp_mode = atoi(optarg);
	    break;

	default:
	    if (opt == ':')
		fprintf(stderr, "Missing parameter\n");
	    else
		fprintf(stderr, "Unknown option '%c'\n", opt);
	    usage();
	    return 1;
	}
    }
    
    if (optind == argc) {
	usage();
	return 1;
    }

    if (NULL == (io = g_open(argv[optind], 1))) {
	fprintf(stderr, "Unable to open db: %s\n", argv[1]);
	return 1;
    }
    optind++;

    if (optind != argc) {
	xpos = atoi(argv[optind]);
    }

    v = lock(io, 0);
    g_read_le4(io, v, &io->c, sizeof(io->c));
    unlock(io, v);

    store->io = io;
    store->get_bin = get_bin;
    store->free_bin = free_bin;
    store->root_bin = io->c.bin;
    store->bin_hash = HashTableCreate(5, HASH_DYNAMIC_SIZE | HASH_OWN_KEYS);
    store->bin_hash->clientdata = store;
    store->bin_hash->load = io_bin_load2;
    store->bin_hash->del  = io_bin_free;

    store->seq_hash = HashTableCreate(5, HASH_DYNAMIC_SIZE | HASH_OWN_KEYS);
    store->seq_hash->clientdata = store;
    store->seq_hash->load = io_seq_load2;

    /* test_mode(store, xpos); */
    /* test_mode2(store, xpos); */

    if (lp_mode) {
	print_output(store, xpos, lp_mode, mode);
    } else {
	init_curses();
	curses_loop(store, xpos, mode);
	endwin();
	store_destroy(store);
    }

    if (!lp_mode) {
	printf("\n\n\tg_view:\tShort Read Alignment Viewer, version 0.3\n");
	printf("\n\tAuthor:\tJames Bonfield (jkb@sanger.ac.uk)\n");
	printf("\t\t2007, Wellcome Trust Sanger Institute\n\n");
    }

    return 0;
}
